# noqa: INP001
# ruff: noqa: S101; flake8: DC102
"""Unit and fuzz tests for ``ozi-fix`` utility script"""
# Part of ozi.
# See LICENSE.txt in the project root for details.
from __future__ import annotations

import argparse
import difflib
import os
import pathlib
from copy import deepcopy
from datetime import timedelta

import pytest
from hypothesis import given
from hypothesis import settings
from hypothesis import strategies as st
from ozi_spec import METADATA  # pyright: ignore
from ozi_templates import load_environment  # pyright: ignore

import ozi_core.fix.missing  # pyright: ignore
import ozi_core.fix.rewrite_command  # pyright: ignore
import ozi_core.pkg_extra  # pyright: ignore
import ozi_core.render  # pyright: ignore
from ozi_core.fix.build_definition import unrollable_subdirs  # pyright: ignore

required_pkg_info_patterns = (
    'Author',
    'Author-email',
    'Description-Content-Type',
    'Home-page',
    'License',
    'License-Expression',
    'License-File',
    'Metadata-Version',
    'Name',
    'Programming Language :: Python',
    'Summary',
    'Version',
)

SAMPLE_MESON_BUILD = """# ozi/meson.build
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
custom_target(
    '_locales.py',
    input: 'generate_locales.py',
    output: '_locales.py',
    command: [python, '@INPUT@'],
    build_by_default: true,
    install: true,
    install_dir: python.get_install_dir() / 'ozi_core',
)
source_files = [
  '__init__.py',
  '__init__.pyi',
  '_i18n.py',
  '_i18n.pyi',
  '_locales.pyi',
  'actions.py',
  'actions.pyi',
  'comment.py',
  'comment.pyi',
  'generate_locales.py',
  'generate_locales.pyi',
  'meson.py',
  'meson.pyi',
  'pkg_extra.py',
  'pkg_extra.pyi',
  'py.typed',
  'render.py',
  'render.pyi',
  'spdx.py',
  'spdx.pyi',
  'trove.py',
  'trove.pyi'
]
foreach file : files(source_files)
    fs.copyfile(file)
    if not meson.is_subproject() or get_option('install-subprojects').enabled()
        python.install_sources(file, pure: true, subdir: 'ozi_core')
    endif
endforeach
source_children = [
    'data',
    'vendor',
    'fix',
    'new',
    'ui',
]
foreach child: source_children
    if child != 'ozi.phony'
        subdir(child)
    endif
endforeach
if false
    executable('source_files', source_files)
    executable('ext_files', ext_files)
    executable('source_children', source_children)
endif
foo_children = []
foreach child: foo_children
    if child != 'ozi.phony'
        subdir(child)
    endif
endforeach
"""

bad_namespace = argparse.Namespace(
    strict=False,
    verify_email=True,
    name='OZI-phony',
    keywords='foo,bar,baz',
    maintainer=[],
    maintainer_email=[],
    author=['foo'],
    author_email=['noreply@oziproject.dev'],
    home_page='https://oziproject.dev',
    summary='A' * 512,
    copyright_head='',
    license_expression='CC0-1.0',
    license_file='LICENSE.txt',
    license='CC0 1.0 Universal (CC0 1.0) Public Domain Dedication',
    license_id='ITWASAFICTION',
    license_exception_id='WEMADEITUP',
    topic=['Utilities'],
    status=['7 - Inactive'],
    environment=['Other Environment'],
    framework=['Pytest'],
    audience=['Other Audience'],
    ci_provider='github',
    project_url=['Home, https://oziproject.dev'],
    long_description_content_type='md',
    fix='',
    add=['ozi.phony'],
    remove=['ozi.phony'],
    dist_requires=[],
    allow_file=[],
    missing=True,
)

env = load_environment(vars(bad_namespace), METADATA.asdict())


@pytest.fixture
def bad_project(tmp_path_factory: pytest.TempPathFactory) -> pathlib.Path:
    """Fixture to wrap the ``ozi-new project`` functionality."""
    fn = tmp_path_factory.mktemp('project_')
    namespace = deepcopy(bad_namespace)
    namespace.target = fn
    preprocessed = ozi_core.new.validate.preprocess_arguments(namespace)
    postprocessed = ozi_core.new.validate.postprocess_arguments(preprocessed)
    ozi_core.render.RenderedContent(
        load_environment(vars(postprocessed), METADATA.asdict()),
        postprocessed.target,
        postprocessed.name,
        postprocessed.ci_provider,
        postprocessed.long_description_content_type,
        True,
    ).render()
    return fn


@pytest.mark.parametrize(
    'key',
    [i for i in METADATA.spec.python.src.required.root if i not in ['PKG-INFO']],
)
def test_report_missing_required_root_file(
    bad_project: pathlib.Path,
    key: str,
) -> None:
    """Check that we warn on missing files."""
    os.remove(bad_project.joinpath(key))
    with pytest.raises(RuntimeWarning):
        ozi_core.fix.missing.report(bad_project)


@pytest.mark.parametrize('key', METADATA.spec.python.src.required.test)
def test_report_missing_required_test_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files."""
    os.remove(bad_project.joinpath('tests') / key)
    with pytest.raises(RuntimeWarning):
        ozi_core.fix.missing.report(bad_project)


@pytest.mark.parametrize('key', METADATA.spec.python.src.required.source)
def test_report_missing_required_source_file(bad_project: pathlib.Path, key: str) -> None:
    """Check that we warn on missing files."""
    os.remove(bad_project.joinpath('ozi_phony') / key)
    with pytest.raises(RuntimeWarning):
        ozi_core.fix.missing.report(bad_project)


@given(
    type=st.just('target'),
    target=st.text(min_size=1, max_size=20),
    operation=st.text(min_size=1, max_size=20),
    sources=st.lists(st.text(min_size=1, max_size=20)),
    subdir=st.just(''),
    target_type=st.just('executable'),
)
def test_fuzz_RewriteCommand(  # noqa: N802, DC102, RUF100
    type: str,  # noqa: A002
    target: str,
    operation: str,
    sources: list[str],
    subdir: str,
    target_type: str,
) -> None:
    ozi_core.fix.rewrite_command.RewriteCommand(
        type=type,
        target=target,
        operation=operation,
        sources=sources,
        subdir=subdir,
        target_type=target_type,
    )


@given(
    target=st.just('.'),
    name=st.text(min_size=1, max_size=20),
    fix=st.sampled_from(('test', 'source', 'root')),
    commands=st.lists(
        st.dictionaries(
            keys=st.text(min_size=1, max_size=20),
            values=st.text(min_size=1, max_size=20),
        ),
    ),
)
def test_fuzz_Rewriter(  # noqa: N802, DC102, RUF100
    target: str,
    name: str,
    fix: str,
    commands: list[dict[str, str]],
) -> None:
    ozi_core.fix.rewrite_command.Rewriter(
        target=target,
        name=name,
        fix=fix,
        commands=commands,
        env=env,
    )


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__dir_nested_warns(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    with pytest.warns(RuntimeWarning):
        rewriter += ['foo/foo/baz/']


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__dir(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    rewriter += ['foo/']
    assert len(rewriter.commands) == 1


def test_Rewriter_bad_project__iadd__bad_fix(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix='',
        env=env,
    )
    with pytest.warns(RuntimeWarning):
        rewriter += ['foo/']
    assert len(rewriter.commands) == 0


def test_Rewriter_bad_project__isub__bad_fix(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix='',
        env=env,
    )
    rewriter -= ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__non_existing_child(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    with pytest.raises(RuntimeWarning):
        rewriter -= ['foo/']
    assert len(rewriter.commands) == 0


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__child(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo').mkdir()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo').mkdir()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo').mkdir()
    rewriter -= ['foo/']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__python_file(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo.py').touch()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo.py').touch()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo.py').touch()
    rewriter -= ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__isub__file(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    if fix == 'root':
        pathlib.Path(str(bad_project), 'foo').touch()
    elif fix == 'source':
        pathlib.Path(str(bad_project), 'ozi_phony', 'foo').touch()
    elif fix == 'test':
        pathlib.Path(str(bad_project), 'tests', 'foo').touch()
    rewriter -= ['foo']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__file(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    rewriter += ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__file_from_template(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest | pathlib.Path,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    pathlib.Path(bad_project / 'templates' / 'foo.py').touch()  # pyright: ignore
    pathlib.Path(bad_project / 'templates' / 'source').mkdir()  # pyright: ignore
    pathlib.Path(bad_project / 'templates' / 'source' / 'foo.py').touch()  # pyright: ignore
    pathlib.Path(bad_project / 'templates' / 'test').mkdir()  # pyright: ignore
    pathlib.Path(bad_project / 'templates' / 'test' / 'foo.py').touch()  # pyright: ignore
    rewriter += ['foo.py']
    assert len(rewriter.commands) == 1


@pytest.mark.parametrize('fix', ['test', 'root', 'source'])
def test_Rewriter_bad_project__iadd__non_python_file(  # noqa: N802, DC102, RUF100
    bad_project: pytest.FixtureRequest,
    fix: str,
) -> None:
    rewriter = ozi_core.fix.rewrite_command.Rewriter(
        target=str(bad_project),
        name='ozi_phony',
        fix=fix,
        env=env,
    )
    with pytest.warns(RuntimeWarning):
        rewriter += ['foo']
    assert len(rewriter.commands) == 1


header = """.. OZI
  Classifier: License-Expression :: Apache-2.0 WITH LLVM-exception
  Classifier: License-File :: LICENSE.txt
"""


@settings(deadline=timedelta(milliseconds=1000))
@given(payload=st.text(max_size=65535).map(header.__add__), as_message=st.booleans())
def test_fuzz_pkg_info_extra(payload: str, as_message: bool) -> None:  # noqa: DC102, RUF100
    ozi_core.pkg_extra._pkg_info_extra(
        payload=payload,
        as_message=as_message,
    )


def test_meson_unroll_subdirs() -> None:
    x = unrollable_subdirs.parse_string(SAMPLE_MESON_BUILD)
    y = '\n'.join([i.rstrip('\n') for i in x])
    assert "\nfoo_children = ['ozi.phony']\n" in y
    assert "\nsource_children = ['ozi.phony']\nforeach" in y
    assert "\nsubdir('data')\n" in y
    assert "\nsubdir('vendor')\n" in y
    assert "\nsubdir('fix')\n" in y
    assert "\nsubdir('new')\n" in y
    assert "\nsubdir('ui')\n" in y
    assert y != SAMPLE_MESON_BUILD
    z = '\n'.join([i.rstrip('\n') for i in unrollable_subdirs.parse_string(y)])
    assert y == z
    assert ''.join(difflib.context_diff(SAMPLE_MESON_BUILD, y)) == ''.join(
        difflib.context_diff(SAMPLE_MESON_BUILD, z),
    )



import os
from configparser import ConfigParser
from pathlib import Path

from ozi_templates.filter import get_ozi_tarball_sha256  # pyright: ignore


def update_wrapfile(target: Path | str, version: str) -> None:  # noqa: C901
    """Update a project :file:`subprojects/ozi.wrap` and symlink to the latest OZI version.

    :param version: release to search for
    :type version: str
    """
    config = ConfigParser()
    ozi_wrap = Path(target, 'subprojects', 'ozi.wrap')
    config.read(ozi_wrap)
    if 'wrap-file' not in config:  # pragma: defer to E2E
        config.add_section('wrap-file')
    if config.remove_section('wrap-git'):  # pragma: defer to E2E
        config.remove_section('provide')
    wrap_file = config['wrap-file']
    wrap_file['directory'] = f'OZI-{version}'
    if 'PYTEST_CURRENT_TEST' not in os.environ:  # pragma: defer to E2E
        wrap_file['source_url'], wrap_file['source_hash'] = get_ozi_tarball_sha256(version)
    else:
        wrap_file['source_url'], wrap_file['source_hash'] = 'pytest', 'pytest'
    wrap_file['source_filename'] = f'OZI-{version}.tar.gz'
    if 'provide' not in config:  # pragma: defer to E2E
        config.add_section('provide')
    provide = config['provide']
    provide['dependency_names'] = f'ozi, ozi-{version}'
    with ozi_wrap.open('w', encoding='utf-8') as f:
        config.write(f)
    if (ozi_wrap.parent / 'ozi').is_symlink():
        (ozi_wrap.parent / 'ozi').unlink()  # pragma: defer to E2E
    (ozi_wrap.parent / 'ozi').symlink_to(
        '..' / ozi_wrap.parent / f'OZI-{version}', target_is_directory=True
    )



# ozi/_i18l.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Internationalization utilities."""
from __future__ import annotations

import locale
from logging import getLogger
from string import Template
from typing import Any

from ozi_core._locales import data
from ozi_core._logging import PytestFilter
from ozi_core._logging import config_logger

config_logger()
_LOCALE = locale.getlocale()[0]


class Translation:
    """Translation API for use inside OZI tools.

    Try to get the system locale and load translations.
    """

    __slots__ = ('__logger', '_locale', 'data')

    def __init__(self: Translation) -> None:
        self.data = data
        self._locale = (
            _LOCALE[:2] if _LOCALE is not None and _LOCALE[:2] in self.data else 'en'
        )
        self.__logger = getLogger(f'ozi_core.{__name__}.{self.__class__.__name__}')
        self.__logger.addFilter(PytestFilter())

    @property
    def locale(self: Translation) -> str | Any:  # pragma: no cover
        """Get the current locale setting."""
        return self._locale

    @locale.setter
    def locale(self: Translation, loc: str) -> None:  # pragma: no cover
        """Set the locale for string translation."""
        if loc in self.data:
            self._locale = loc
        else:
            self.__logger.debug(f'Invalid locale: {loc}')

    def __call__(self: Translation, _key: str, **kwargs: str) -> str:  # pragma: no cover
        """Get translation text by key and pass optional substitions as keyword arguments."""
        if self.locale not in self.data:
            return _key
        text = self.data[self.locale].get(_key, _key)
        if text is None:
            return ''
        elif text is _key:
            self.__logger.debug(f'no translation for "{_key}" in locale "{self.locale}"')
        return Template(text).safe_substitute(**kwargs)


TRANSLATION = Translation()



# ozi/fix/__init__.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""ozi-fix: Project fix script that outputs a meson rewriter JSON array."""
from __future__ import annotations

import json
import os
import sys
from contextlib import suppress
from pathlib import Path
from subprocess import PIPE
from subprocess import Popen
from typing import TYPE_CHECKING
from typing import NoReturn
from unittest.mock import Mock

if sys.platform != 'win32':
    import termios
    import tty
else:  # pragma: no cover
    tty = Mock()
    termios = Mock()
    tty.setraw = lambda x: None
    termios.tcgetattr = lambda x: None
    termios.tcsetattr = lambda x, y, z: None

from ozi_spec import METADATA
from ozi_templates import load_environment
from ozi_templates.filter import underscorify  # pyright: ignore
from tap_producer import TAP

from ozi_core.fix.build_definition import unroll_subdirs
from ozi_core.fix.build_definition import walk
from ozi_core.fix.interactive import interactive_prompt
from ozi_core.fix.missing import get_relpath_expected_files
from ozi_core.fix.missing import report
from ozi_core.fix.parser import parser
from ozi_core.fix.rewrite_command import Rewriter
from ozi_core.fix.wrap import update_wrapfile
from ozi_core.new.validate import valid_copyright_head

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace

    from jinja2 import Environment


def _setup(project: Namespace) -> tuple[Namespace, Environment]:  # pragma: no cover
    TAP.version(14)
    project.target = Path(os.path.relpath(os.path.join('/', project.target), '/')).absolute()
    if not project.target.exists():
        TAP.bail_out(f'target: {project.target} does not exist.')
    elif not project.target.is_dir():
        TAP.bail_out(f'target: {project.target} is not a directory.')
    with suppress(ValueError):
        project.add.remove('ozi.phony')
        project.remove.remove('ozi.phony')
    project.add = list(set(project.add))
    project.remove = list(set(project.remove))
    env = load_environment(vars(project), METADATA.asdict())  # pyright: ignore
    return project, env


def main(args: list[str] | None = None) -> NoReturn:  # pragma: no cover
    """Main ozi.fix entrypoint."""
    project = parser.parse_args(args=args)
    project.missing = project.fix in {'missing', 'm', 'mis'}
    project.interactive = project.fix in {'interactive', 'i'}
    if project.update_wrapfile:
        update_wrapfile(project.target, METADATA.ozi.version)
    match [project.interactive, project.missing, project.strict]:
        case [True, False, _]:
            with TAP.suppress():  # pyright: ignore
                project, env = _setup(project)
                name, *_ = report(project.target)
                project.name = underscorify(name)
            fd = sys.stdin.fileno()
            original_attributes = termios.tcgetattr(fd)
            tty.setraw(sys.stdin)
            args = interactive_prompt(project)
            termios.tcsetattr(fd, termios.TCSADRAIN, original_attributes)
            TAP.comment(f'ozi-fix {" ".join(args)}')
            main(args)
        case [False, True, False]:
            project, _ = _setup(project)
            name, *_ = report(project.target)
        case [False, False, _]:
            with TAP.suppress():  # pyright: ignore
                project, env = _setup(project)
                name, *_ = report(project.target)
                project.name = underscorify(name)
                project.license_file = 'LICENSE.txt'
                project.copyright_head = valid_copyright_head(
                    project.copyright_head,
                    name,
                    project.license_file,
                )
                rewriter = Rewriter(str(project.target), project.name, project.fix, env)
                rewriter += project.add
                rewriter -= project.remove
                for d in walk(
                    project.target,
                    get_relpath_expected_files(project.fix, name)[0],
                    [],
                ):
                    for k in d.keys():
                        build_text = unroll_subdirs(project.target, k)
                        Path(project.target / k / 'meson.build').write_text(build_text)
                TAP.plan()
            if len(project.add) > 0 or len(project.remove) > 0:
                out = json.dumps(rewriter.commands, indent=4 if project.pretty else None)
                if project.interactive_io:
                    res = Popen(
                        ['tox', '-e', 'invoke', '--', 'rewrite', out],
                        stdin=PIPE,
                    )
                    res.communicate()
                    if res.returncode != 0:
                        TAP.comment('ozi-fix failed to rewrite project files')
                    exit(0)
                else:
                    print(out)
                    exit(0)
            else:
                parser.print_help()
        case [False, True, True]:
            with TAP.strict():  # pyright: ignore
                project, _ = _setup(project)
                name, *_ = report(project.target)
        case [True, True, _]:
            TAP.bail_out('subcommands `interactive` and `missing` are mutually exclusive')
        case [_, _, _]:
            TAP.bail_out('Name discovery failed.')
    TAP.end()
    exit(0)



# ozi/spdx.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""SPDX standard metadata parser expression grammars, with support for :pep:`639` keys."""
from ozi_spec import METADATA
from pyparsing import Combine
from pyparsing import Forward
from pyparsing import Keyword
from pyparsing import Literal
from pyparsing import MatchFirst
from pyparsing import Optional
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import alphanums
from pyparsing import oneOf
from spdx_license_list import LICENSES

user_defined_license = Combine(
    Optional('DocumentRef-' + Word(alphanums + '-.') + ':')
    + 'LicenseRef-'
    + Word(alphanums + '-.'),
)
spdx_license_expression = Forward()
spdx_license_expression <<= MatchFirst(
    [
        user_defined_license,
    ]
    + [Literal(lic.id) for lic in LICENSES.values() if not lic.deprecated_id],
).set_name('License-ID') + ZeroOrMore(
    Keyword('WITH')
    + oneOf(METADATA.spec.python.pkg.license.exceptions).set_name('License-Exception-ID')
    | Keyword('AND') + spdx_license_expression
    | Keyword('OR') + spdx_license_expression,
) | Literal(
    '(',
) + spdx_license_expression + Literal(
    ')',
)



"""
This type stub file was generated by pyright.
"""

from typing import Optional
from typing import Union

from .exceptions_types import ValidatedEmail

def validate_email(email: Union[str, bytes], /, *, allow_smtputf8: Optional[bool] = ..., allow_empty_local: bool = ..., allow_quoted_local: Optional[bool] = ..., allow_domain_literal: Optional[bool] = ..., check_deliverability: Optional[bool] = ..., test_environment: Optional[bool] = ..., globally_deliverable: Optional[bool] = ..., timeout: Optional[int] = ..., dns_resolver: Optional[object] = ...) -> ValidatedEmail:
    """
    Given an email address, and some options, returns a ValidatedEmail instance
    with information about the address if it is valid or, if the address is not
    valid, raises an EmailNotValidError. This is the main function of the module.
    """




"""
This type stub file was generated by pyright.
"""

from .exceptions_types import EmailNotValidError
from .exceptions_types import EmailSyntaxError
from .exceptions_types import EmailUndeliverableError
from .exceptions_types import ValidatedEmail
from .validate_email import validate_email
from .version import __version__

__all__ = ['validate_email', 'ValidatedEmail', 'EmailNotValidError', 'EmailSyntaxError', 'EmailUndeliverableError', 'caching_resolver', '__version__']
def caching_resolver(*args, **kwargs): # -> Resolver:
    ...

ALLOW_SMTPUTF8 = ...
ALLOW_QUOTED_LOCAL = ...
ALLOW_DOMAIN_LITERAL = ...
GLOBALLY_DELIVERABLE = ...
CHECK_DELIVERABILITY = ...
TEST_ENVIRONMENT = ...
DEFAULT_TIMEOUT = ...
SPECIAL_USE_DOMAIN_NAMES = ...



"""
This type stub file was generated by pyright.
"""

def split_email(email): # -> tuple[str, str | Any, Literal[True]] | tuple[Any, Any, Literal[False]]:
    ...

def get_length_reason(addr, utf8=..., limit=...) -> str:
    """Helper function to return an error message related to invalid length."""

def safe_character_display(c): # -> LiteralString | str:
    ...

def validate_email_local_part(local: str, allow_smtputf8: bool = ..., allow_empty_local: bool = ..., quoted_local_part: bool = ...): # -> dict[str, Any]:
    """Validates the syntax of the local part of an email address."""

def check_unsafe_chars(s, allow_space=...): # -> None:
    ...

def check_dot_atom(label, start_descr, end_descr, is_hostname): # -> None:
    ...

def validate_email_domain_name(domain, test_environment=..., globally_deliverable=...): # -> dict[str, str]:
    """Validates the syntax of the domain part of an email address."""

def validate_email_length(addrinfo): # -> None:
    ...

def validate_email_domain_literal(domain_literal): # -> dict[str, Any]:
    ...




from __future__ import annotations

import os
import sys
from contextlib import suppress
from dataclasses import asdict
from io import UnsupportedOperation
from typing import TYPE_CHECKING
from typing import Any
from unittest.mock import Mock

from prompt_toolkit.shortcuts.dialogs import button_dialog
from prompt_toolkit.shortcuts.dialogs import checkboxlist_dialog
from prompt_toolkit.shortcuts.dialogs import message_dialog
from prompt_toolkit.shortcuts.dialogs import radiolist_dialog
from prompt_toolkit.shortcuts.dialogs import yes_no_dialog
from tap_producer import TAP

from ozi_core._i18n import TRANSLATION
from ozi_core.config import OziConfig
from ozi_core.config import read_user_config
from ozi_core.config import write_user_config
from ozi_core.fix.build_definition import walk
from ozi_core.fix.missing import get_relpath_expected_files
from ozi_core.fix.validate import RewriteCommandTargetValidator
from ozi_core.new.interactive.validator import validate_message
from ozi_core.ui._style import _style
from ozi_core.ui.defaults import COPYRIGHT_HEAD
from ozi_core.ui.defaults import FIX_PRETTY
from ozi_core.ui.defaults import FIX_STRICT
from ozi_core.ui.dialog import input_dialog
from ozi_core.ui.menu import MenuButton
from ozi_core.ui.menu import checkbox

if sys.platform != 'win32':  # pragma: no cover
    import curses
else:  # pragma: no cover
    curses = Mock()
    curses.tigetstr = lambda x: b''
    curses.setupterm = lambda: None

if TYPE_CHECKING:  # pragma: no cover
    from argparse import Namespace
    from pathlib import Path


def options_menu(  # pragma: no cover
    prompt: Any,
    output: dict[str, list[str]],
    prefix: dict[str, str],
) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    _default: str | list[str] | None = None
    while True:
        match radiolist_dialog(
            title=TRANSLATION('fix-dlg-title'),
            text=TRANSLATION('opt-menu-title'),
            values=[
                (
                    'strict',
                    TRANSLATION('opt-menu-strict', value=checkbox(prompt.strict)),
                ),
                (
                    'pretty',
                    TRANSLATION('opt-menu-pretty', value=checkbox(prompt.pretty)),
                ),
                ('copyright_head', TRANSLATION('opt-menu-copyright-head')),
                (
                    'language',
                    TRANSLATION(
                        'opt-menu-language',
                        value=TRANSLATION(f'lang-{TRANSLATION.locale}'),
                    ),
                ),
            ],
            style=_style,
            cancel_text=MenuButton.BACK._str,
            ok_text=MenuButton.OK._str,
        ).run():
            case x if x and x in ('strict', 'pretty'):
                for i in (f'--{x}', f'--no-{x}'):
                    if i in output:
                        output.pop(i)
                setting = getattr(prompt, x)
                if setting is None:
                    setattr(prompt, x, False)
                else:
                    flag = '' if setting else 'no-'
                    output.update(
                        {
                            f'--{flag}{x.replace("_", "-")}': [
                                f'--{flag}{x.replace("_", "-")}',
                            ],
                        },
                    )
                    setattr(prompt, x, not setting)
            case x if x and x == 'copyright_head':
                _default = output.setdefault('--copyright-head', [COPYRIGHT_HEAD])
                result = input_dialog(
                    title=TRANSLATION('fix-dlg-title'),
                    text=TRANSLATION('opt-menu-copyright-head-input'),
                    style=_style,
                    cancel_text=MenuButton.BACK._str,
                    ok_text=MenuButton.OK._str,
                    default=_default[0],
                    multiline=True,
                ).run()
                if result in _default:
                    prompt.copyright_head = result
                    output.update({'--copyright-head': [prompt.copyright_head]})
            case x if x == 'language':
                result = radiolist_dialog(
                    title=TRANSLATION('fix-dlg-title'),
                    text=TRANSLATION('opt-menu-language-text'),
                    values=list(
                        zip(
                            TRANSLATION.data.keys(),  # type: ignore
                            [TRANSLATION(f'lang-{i}') for i in TRANSLATION.data.keys()],  # type: ignore
                        ),
                    ),
                    cancel_text=MenuButton.BACK._str,
                    ok_text=MenuButton.OK._str,
                    default=TRANSLATION.locale,
                    style=_style,
                ).run()
                if result is not None:
                    TRANSLATION.locale = result
            case _:
                if yes_no_dialog(
                    title=TRANSLATION('fix-dlg-title'),
                    text=TRANSLATION('opt-menu-save-config'),
                    yes_text=MenuButton.YES._str,
                    no_text=MenuButton.NO._str,
                    style=_style,
                ).run():
                    config = asdict(read_user_config())
                    config['fix'].update(
                        **{
                            k: v
                            for k, v in vars(prompt).items()
                            if k not in ['fix', 'target']
                        }
                    )
                    config['interactive'].update(language=TRANSLATION.locale)
                    write_user_config(OziConfig(**config))
                break
    return None, output, prefix


def main_menu(  # pragma: no cover
    prompt: Any,
    output: dict[str, list[str]],
    prefix: dict[str, str],
) -> tuple[None | list[str] | bool, dict[str, list[str]], dict[str, str]]:
    while True:
        match button_dialog(
            title=TRANSLATION('new-dlg-title'),
            text=TRANSLATION('main-menu-text'),
            buttons=[
                MenuButton.OPTIONS._tuple,
                MenuButton.RESET._tuple,
                MenuButton.EXIT._tuple,
                MenuButton.BACK._tuple,
            ],
            style=_style,
        ).run():
            case MenuButton.OPTIONS.value:
                result, output, prefix = options_menu(prompt, output, prefix)
                if isinstance(result, list):
                    return result, output, prefix
            case MenuButton.BACK.value:
                break
            case MenuButton.RESET.value:
                if yes_no_dialog(
                    title=TRANSLATION('new-dlg-title'),
                    text=TRANSLATION('main-menu-yn-reset'),
                    style=_style,
                    yes_text=MenuButton.YES._str,
                    no_text=MenuButton.NO._str,
                ).run():
                    return ['interactive', '.'], output, prefix
            case MenuButton.EXIT.value:
                if yes_no_dialog(
                    title=TRANSLATION('new-dlg-title'),
                    text=TRANSLATION('main-menu-yn-exit'),
                    style=_style,
                    yes_text=MenuButton.YES._str,
                    no_text=MenuButton.NO._str,
                ).run():
                    return ['-h'], output, prefix
    return None, output, prefix


class Prompt:
    def __init__(  # pragma: no cover
        self: Prompt,
        target: Path,
        strict: bool | None = None,
        pretty: bool | None = None,
        copyright_head: str | None = None,
    ) -> None:
        self.target = target
        self.fix: str | None = 'root'
        self.strict = strict if strict is not None else FIX_STRICT
        self.pretty = pretty if pretty is not None else FIX_PRETTY
        self.copyright_head = (
            copyright_head if copyright_head is not None else COPYRIGHT_HEAD
        )
        TRANSLATION.locale = asdict(read_user_config())['interactive']['language']

    def set_fix_mode(  # pragma: no cover
        self: Prompt,
        project_name: str,
        output: dict[str, list[str]],
        prefix: dict[str, str],
    ) -> tuple[list[str] | str | bool | None, dict[str, list[str]], dict[str, str]]:
        while True:
            self.fix = radiolist_dialog(
                title=TRANSLATION('fix-dlg-title'),
                text=TRANSLATION('fix-add'),
                style=_style,
                cancel_text=MenuButton.MENU._str,
                ok_text=MenuButton.OK._str,
                values=[('source', 'source'), ('test', 'test'), ('root', 'root')],
            ).run()
            if self.fix is not None:
                output['fix'].append(self.fix)
                return None, output, prefix
            else:
                result, output, prefix = main_menu(self, output, prefix)
                if result is not None:
                    return result, output, prefix

    def add_or_remove(  # pragma: no cover
        self: Prompt,
        project_name: str,
        output: dict[str, list[str]],
        prefix: dict[str, str],
    ) -> tuple[list[str] | str | bool | None, dict[str, list[str]], dict[str, str]]:
        add_files: list[str] = []
        rem_files: list[str] = []
        output.setdefault('--add', [])
        output.setdefault('--remove', [])
        while True:
            match button_dialog(
                title=TRANSLATION('fix-dlg-title'),
                text='\n'.join(
                    (
                        '\n'.join(iter(prefix)),
                        '\n',
                        TRANSLATION('fix-add-or-remove', projectname=project_name),
                    ),
                ),
                buttons=[
                    MenuButton.ADD._tuple,
                    MenuButton.REMOVE._tuple,
                    MenuButton.MENU._tuple,
                    MenuButton.OK._tuple,
                ],
                style=_style,
            ).run():
                case MenuButton.ADD.value:
                    rel_path, _ = get_relpath_expected_files(self.fix, project_name)
                    files = []
                    with TAP.suppress():
                        for d in walk(self.target, rel_path, []):
                            for v in d.values():
                                files += [str(i) for i in v['missing']]
                    result = checkboxlist_dialog(
                        title=TRANSLATION('fix-dlg-title'),
                        text='',
                        values=[('input', '<input>')] + [(i, i) for i in sorted(files)],
                        style=_style,
                    ).run()
                    if result is not None:
                        add_files += [i for i in result if i != 'input']
                        if len(add_files) > 0:
                            prefix.update(
                                {
                                    f'Add-{self.fix}: {add_files}': (
                                        f'Add-{self.fix}: {add_files}'
                                    ),
                                },
                            )
                            for f in add_files:
                                output['--add'].append(f)
                        if 'input' in set(result):
                            result = input_dialog(
                                title=TRANSLATION('fix-dlg-title'),
                                cancel_text=MenuButton.MENU._str,
                                style=_style,
                                validator=RewriteCommandTargetValidator(),
                            ).run()
                            if result is not None:
                                valid, errmsg = validate_message(
                                    result,
                                    RewriteCommandTargetValidator(),
                                )
                                if valid:
                                    add_files += [result]
                                    prefix.update(
                                        {
                                            f'Add-{self.fix}: {add_files}': (
                                                f'Add-{self.fix}: {add_files}'
                                            ),
                                        },
                                    )
                                    output['--add'].append(str(result))
                                else:
                                    message_dialog(
                                        title=TRANSLATION('fix-dlg-title'),
                                        text=TRANSLATION(
                                            'msg-input-invalid',
                                            value=result,
                                            errmsg=errmsg,
                                        ),
                                        style=_style,
                                        ok_text=MenuButton.OK._str,
                                    ).run()
                case MenuButton.REMOVE.value:
                    rel_path, _ = get_relpath_expected_files(self.fix, project_name)
                    files = []
                    with TAP.suppress():
                        for d in walk(self.target, rel_path, []):
                            for v in d.values():
                                files += [str(i) for i in v['found']]
                    result = checkboxlist_dialog(
                        title=TRANSLATION('fix-dlg-title'),
                        text='',
                        values=[('input', '<input>')] + [(i, i) for i in sorted(files)],
                        style=_style,
                    ).run()
                    if result is not None:
                        rem_files += [i for i in result if i != 'input']
                        if len(rem_files) > 0:
                            prefix.update(
                                {
                                    f'Remove-{self.fix}: {rem_files}': (
                                        f'Remove-{self.fix}: {rem_files}'
                                    ),
                                },
                            )
                            for f in rem_files:
                                output['--remove'].append(f)
                        if 'input' in set(result):
                            result = input_dialog(
                                title=TRANSLATION('fix-dlg-title'),
                                cancel_text=MenuButton.MENU._str,
                                style=_style,
                                validator=RewriteCommandTargetValidator(),
                            ).run()
                            if result is not None:
                                valid, errmsg = validate_message(
                                    result,
                                    RewriteCommandTargetValidator(),
                                )
                                if valid:
                                    rem_files += [result]
                                    prefix.update(
                                        {
                                            f'Remove-{self.fix}: {rem_files}': (
                                                f'Remove-{self.fix}: {rem_files}'
                                            ),
                                        },
                                    )
                                    output['--remove'].append(str(result))
                                else:
                                    message_dialog(
                                        title=TRANSLATION('fix-dlg-title'),
                                        text=TRANSLATION(
                                            'msg-input-invalid',
                                            value=result,
                                            errmsg=errmsg,
                                        ),
                                        style=_style,
                                        ok_text=MenuButton.OK._str,
                                    ).run()
                case MenuButton.OK.value:
                    break
                case MenuButton.MENU.value:
                    result, output, prefix = main_menu(self, output, prefix)
                    if result is not None:
                        return result, output, prefix
        return None, output, prefix


def interactive_prompt(project: Namespace) -> list[str]:  # pragma: no cover # noqa: C901
    ret_args = []
    with suppress(UnsupportedOperation):
        curses.setupterm()
        e3 = curses.tigetstr('E3') or b''
        clear_screen_seq = curses.tigetstr('clear') or b''
        os.write(sys.stdout.fileno(), e3 + clear_screen_seq)
    p = Prompt(project.target, **asdict(read_user_config())['fix'])
    result, output, prefix = p.set_fix_mode(
        project_name=project.name,
        output={'fix': []},
        prefix={},
    )
    if isinstance(result, list):
        return result
    result, output, prefix = p.add_or_remove(
        project_name=project.name,
        output=output,
        prefix=prefix,
    )
    if isinstance(result, list):
        return result
    fix = output.pop('fix')
    for k, v in output.items():
        for i in v:
            if len(i) > 0:
                ret_args += [k, i]
    return fix + ret_args + ['--interactive-io', '.']



from dataclasses import asdict
from dataclasses import dataclass
from dataclasses import field
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version
from pathlib import Path

import yaml
from platformdirs import user_config_dir

from ozi_core import __version__

try:
    core_version = version('ozi-core')
except PackageNotFoundError:
    core_version = 'git-dev'


HEADER = '\n'.join(
    [
        f'# {Path(user_config_dir("OZI")) / "config.yml"}',
        f'# OZI version: {__version__}',
        f'# ozi-core version: {core_version}\n',
    ]
)
CONF_PATH = Path(user_config_dir('OZI')) / 'config.yml'


@dataclass(kw_only=True)
class OziInteractiveConfig:
    """General config options for dialog-based CLI."""

    language: str | None = None


@dataclass(kw_only=True)
class OziFixConfig:
    """Persistent ``ozi-fix interactive`` settings."""

    copyright_head: str | None = None
    pretty: bool | None = None
    strict: bool | None = None


@dataclass(kw_only=True)
class OziNewConfig:
    """Persistent ``ozi-new interactive`` settings."""

    allow_file: list[str] | None = None
    author: str | None = None
    author_email: str | None = None
    ci_provider: str | None = None
    check_package_exists: bool | None = None
    copyright_head: str | None = None
    enable_cython: bool | None = None
    enable_uv: bool | None = None
    github_harden_runner: bool | None = None
    maintainer: str | None = None
    maintainer_email: str | None = None
    language: list[str] | None = None
    readme_type: str | None = None
    strict: bool | None = None
    verify_email: bool | None = None


@dataclass(kw_only=True)
class OziConfig:
    """Persistent ``ozi-* interactive`` settings."""

    fix: OziFixConfig = field(default_factory=OziFixConfig)
    new: OziNewConfig = field(default_factory=OziNewConfig)
    interactive: OziInteractiveConfig = field(default_factory=OziInteractiveConfig)


def read_user_config() -> OziConfig:  # pragma: defer to E2E
    CONF_PATH.parent.mkdir(exist_ok=True)
    CONF_PATH.touch(exist_ok=True)
    data = yaml.safe_load(CONF_PATH.read_text())
    if data is not None:
        return OziConfig(**data)
    else:
        return OziConfig()


def write_user_config(  # pragma: defer to E2E
    data: OziConfig,
) -> None:
    CONF_PATH.write_text(
        HEADER + yaml.safe_dump(data=asdict(data), allow_unicode=True),
        encoding='utf-8',
    )



"""Validators for interactive prompts."""

from __future__ import annotations

import re
from functools import lru_cache

import niquests  # pyright: ignore
from prompt_toolkit.document import Document  # pyright: ignore
from prompt_toolkit.validation import ThreadedValidator  # pyright: ignore
from prompt_toolkit.validation import ValidationError  # pyright: ignore
from prompt_toolkit.validation import Validator

from ozi_core._i18n import TRANSLATION


@lru_cache
def pypi_package_exists(package: str) -> bool:  # pragma: no cover
    """Check if a package name exists on PyPI."""
    return (
        niquests.get(
            f'https://pypi.org/simple/{package}',
            timeout=15,
        ).status_code
        == 200
    )


class ProjectNameValidator(Validator):
    """Validate that a package name is valid."""

    def validate(
        self,  # noqa: ANN101,RUF100
        document: Document,
    ) -> None:  # pragma: no cover
        if len(document.text) == 0:
            raise ValidationError(0, TRANSLATION('err-no-empty'))
        if not re.match(
            '^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$',
            document.text,
            flags=re.IGNORECASE,
        ):
            raise ValidationError(len(document.text), TRANSLATION('err-name-invalid'))


class NotReservedValidator(ThreadedValidator):
    """Validate that a package name is available."""

    def validate(
        self,  # noqa: ANN101,RUF100
        document: Document,
    ) -> None:  # pragma: no cover
        self.validator.validate(document)
        if pypi_package_exists(document.text):
            raise ValidationError(len(document.text), TRANSLATION('err-name-exists'))


class LengthValidator(Validator):
    """Validate text is between 1 and 512 chartacters in length."""

    def validate(
        self,  # noqa: ANN101,RUF100
        document: Document,
    ) -> None:  # pragma: no cover
        if len(document.text) == 0:
            raise ValidationError(0, TRANSLATION('err-no-empty'))
        if len(document.text) > 512:
            raise ValidationError(512, TRANSLATION('err-too-long'))


class PackageValidator(Validator):
    """Validate a package name exists on PyPI."""

    def validate(
        self,  # noqa: ANN101,RUF100
        document: Document,
    ) -> None:  # pragma: no cover
        if len(document.text) == 0:
            raise ValidationError(0, TRANSLATION('err-no-empty'))
        if pypi_package_exists(document.text):
            pass
        else:
            raise ValidationError(len(document.text), TRANSLATION('err-pkg-not-found'))


def validate_message(
    text: str,
    validator: Validator,
) -> tuple[bool, str]:  # pragma: no cover
    """Validate a string.

    :param text: string to validate
    :type text: str
    :param validator: validator instance
    :type validator: Validator
    :return: validation, error message
    :rtype: tuple[bool, str]
    """
    try:
        validator.validate(Document(text))
    except ValidationError as e:
        return False, e.message
    return True, ''



# ozi/fix/build_definition.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""Build definition check utilities."""
from __future__ import annotations

import os
from pathlib import Path
from typing import Generator

from ozi_spec import METADATA
from pyparsing import Combine
from pyparsing import DelimitedList
from pyparsing import FollowedBy
from pyparsing import Keyword
from pyparsing import LineEnd
from pyparsing import Literal
from pyparsing import OneOrMore
from pyparsing import Optional
from pyparsing import SkipTo
from pyparsing import StringEnd
from pyparsing import White
from pyparsing import Word
from pyparsing import ZeroOrMore
from pyparsing import alphanums
from pyparsing import alphas
from pyparsing import match_previous_literal
from pyparsing import quoted_string
from pyparsing import rest_of_line
from tap_producer import TAP

from ozi_core import comment
from ozi_core._i18n import TRANSLATION
from ozi_core.meson import get_items_by_suffix
from ozi_core.meson import query_build_value

IGNORE_MISSING = {
    'subprojects',
    *METADATA.spec.python.src.repo.hidden_dirs,
    *METADATA.spec.python.src.repo.ignore_dirs,
    *METADATA.spec.python.src.allow_files,
}
_commasepitem = (
    Combine(
        OneOrMore(
            ~Literal(',')
            + ~LineEnd()
            + quoted_string
            + Optional(White(' \t') + ~FollowedBy(LineEnd() | ',')),
        ),
    )
    .streamline()
    .set_name('commaItem')
)
comma_separated_list = DelimitedList(
    Optional(quoted_string.copy() | _commasepitem, default=''),
).set_name('comma separated list')
SP = White(' ', min=1)
NL = White('\n', min=1)
bound = Word(alphas, alphanums + '_')
assigned = Word(alphas, alphanums + '_')
array_assign = (
    assigned
    + Literal('=')
    + Literal('[')
    + Optional(comma_separated_list, default='')
    + Literal(']')
    + rest_of_line
).set_parse_action(
    lambda t: [
        *[f'subdir({i})' for i in t[3:-2] if len(i) and i.strip('\'"') != 'ozi.phony'],
        ' '.join(t[:3]) + "'ozi.phony'" + t[-2] + t[-1],  # type: ignore
    ],
)
foreach_bind = (
    Keyword('foreach') + bound + ':' + match_previous_literal(assigned) + rest_of_line
).set_parse_action(lambda t: ' '.join(t).strip())
bindvar = match_previous_literal(bound)
foreach_end = (Keyword('endforeach') + rest_of_line).set_parse_action(
    lambda t: ' '.join(t).strip(),
)
if_ozi_phony = (
    Keyword('if') + bindvar + Literal('!=') + Literal("'ozi.phony'") + rest_of_line
).set_parse_action(lambda t: '    ' + ' '.join(t).strip())
endif = (Keyword('endif') + rest_of_line).set_parse_action(
    lambda t: '    ' + ' '.join(t).strip(),
)
subdir_call = ('subdir(' + bindvar + ')' + rest_of_line).set_parse_action(
    lambda t: '        ' + ''.join(t),
)
literal_subdir_loop = (
    array_assign + foreach_bind + if_ozi_phony + subdir_call + endif + foreach_end
)
unrollable_subdirs = (
    ZeroOrMore(SkipTo(literal_subdir_loop, include=True))
    + SkipTo(StringEnd(), include=True).leave_whitespace()
)


def unroll_subdirs(target: Path, rel_path: Path) -> str:  # pragma: defer to E2E
    """Opens a meson.build file and returns the file with literal subdir loops converted
    to single static assignment form.
    """
    with open(target / rel_path / 'meson.build', 'r') as f:
        text = unrollable_subdirs.parse_file(f)
    return '\n'.join([i.strip('\n') for i in text])


def inspect_files(
    target: Path,
    rel_path: Path,
    found_files: list[str],
    extra_files: list[str],
) -> dict[str, list[str]]:  # pragma: no cover
    build_files = [str(rel_path / 'meson.build'), str(rel_path / 'meson.options')]
    _found_files = {'found': [], 'missing': []}
    for file in extra_files:
        found_literal = query_build_value(
            str(target / rel_path),
            file,
        )
        if found_literal and file not in _found_files:
            build_file = str((rel_path / file).parent / 'meson.build')
            TAP.ok(f'{build_file} {TRANSLATION("term-found")} {rel_path / file}')
            build_files += [str(rel_path / file)]
            comment.comment_diagnostic(target, rel_path, file)
            _found_files['found'].append(file)
        if str(rel_path / file) not in build_files and file not in found_files:
            build_file = str(rel_path / 'meson.build')
            TAP.not_ok(f'{build_file} {TRANSLATION("term-missing")} {rel_path / file}')
            _found_files['missing'].append(file)
    return _found_files


def process(
    target: Path,
    rel_path: Path,
    found_files: list[str] | None = None,
) -> dict[str, list[str]]:  # pragma: no cover
    """Process an OZI project build definition's files."""
    try:
        extra_files = [
            file
            for file in os.listdir(target / rel_path)
            if os.path.isfile(target / rel_path / file)
            and not os.path.islink(target / rel_path / file)
        ]
    except FileNotFoundError as e:
        TAP.not_ok(TRANSLATION('term-missing'), e.filename)
        extra_files = []
    found_files = found_files if found_files else []
    extra_files = list(set(extra_files).symmetric_difference(set(found_files)))
    files = inspect_files(
        target=target,
        rel_path=rel_path,
        found_files=found_files,
        extra_files=extra_files,
    )
    return files


def validate(
    target: Path,
    rel_path: Path,
    subdirs: list[str],
    children: set[str] | None,
) -> Generator[Path, None, None]:  # pragma: no cover
    """Validate an OZI standard build definition's directories."""
    for directory in subdirs:
        if directory not in IGNORE_MISSING:
            TAP.ok(
                str(rel_path / 'meson.build'),
                TRANSLATION('term-subdir'),
                str(directory),
            )
            yield Path(rel_path / directory)
        else:
            TAP.ok(
                str(rel_path / 'meson.build'),
                TRANSLATION('term-missing'),
                str(directory),
                skip=True,
            )


def walk(
    target: Path,
    rel_path: Path,
    found_files: list[str] | None = None,
    project_name: str | None = None,
) -> Generator[dict[Path, dict[str, list[str]]], None, None]:  # pragma: no cover
    """Walk an OZI standard build definition directory."""
    files = process(target, rel_path, found_files)
    yield {rel_path: files}
    found_files = files['found'] + files['missing']
    children = list(
        validate(
            target,
            rel_path,
            subdirs=[
                directory
                for directory in os.listdir(target / rel_path)
                if os.path.isdir(target / rel_path / directory)
                and directory not in [project_name, 'tests']
                and rel_path != Path('.')
            ],
            children=get_items_by_suffix(str((target / rel_path)), 'children'),
        ),
    )
    if rel_path == Path('.') and project_name:
        children += [Path('.')]  # pragma: no cover
    for child in children:
        walk(target, child, found_files=found_files)  # pragma: no cover



from __future__ import annotations

import gzip
import json
import logging
import logging.handlers
import os
import sys
from pathlib import Path

from platformdirs import user_log_dir

LOG_PATH = Path(user_log_dir('OZI')) / 'log.json'
LOG_PATH.parent.mkdir(parents=True, exist_ok=True)


class PytestFilter(logging.Filter):
    def filter(self: PytestFilter, record: logging.LogRecord) -> bool:  # pragma: no cover
        return os.environ.get('PYTEST_VERSION') is None or 'pytest' not in sys.modules


class CompressedRotatingFileHandler(logging.handlers.RotatingFileHandler):
    def doRollover(self: CompressedRotatingFileHandler) -> None:  # pragma: no cover
        super().doRollover()
        with gzip.open(f'{self.baseFilename}.gz', 'wb') as f:
            f.write(Path(self.baseFilename).read_bytes())
        Path(self.baseFilename).write_text('')


def config_logger() -> None:
    logger = logging.getLogger('ozi_core')
    logger.propagate = False
    logger.setLevel(logging.DEBUG)
    formatter = logging.Formatter(
        json.dumps(
            {
                'time': '%(asctime)-s',
                'level': '%(levelname)-s',
                'logger': '%(name)s',
                'module': '%(module)-s',
                'funcName': '%(funcName)-s',
                'message': '%(message)s',
            }
        )
        + ','
    )
    handler = CompressedRotatingFileHandler(
        LOG_PATH,
        maxBytes=1_000_000,
        backupCount=5,
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.addFilter(PytestFilter())



"""
This type stub file was generated by pyright.
"""

import re
from collections import Counter
from enum import IntFlag
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Generator
from typing import Sequence

"""Linter comment check utilities."""
if TYPE_CHECKING:
    ...
TIER3_COMMENTS = ...
TIER2_COMMENTS = ...
TIER1_COMMENTS = ...

class CommentQuality(IntFlag):
    """Comment tiers for scoring project quality."""

    TIER1 = ...
    TIER2 = ...
    TIER3 = ...

def calculate_score(lines: int, t1: int, t2: int, t3: int) -> float:
    """Calculate a quality score out of five.
    Comments have more impact on the score when lines is higher.

    :param lines: total line count
    :type lines: int
    :param t1: low-impact comments
    :type t1: int
    :param t2: intermediate-impact comments
    :type t2: int
    :param t3: high-impact comments
    :type t3: int
    :return: comment quality score out of 5.0
    :rtype: float
    """

@lru_cache
def pattern_cache(key: str) -> re.Pattern[str]:
    """Cached OZI specification linter comment pattern lookup.

    :param key: key in :py:class:`ozi_spec.CommentPatterns`
    :type key: str
    :return: compiled regular expression pattern
    :rtype: re.Pattern[str]
    """

def pattern_search(line: str) -> Generator[tuple[str, str], None, None]:
    """Search for OZI specification comment patterns.

    :param line: line text verbatim
    :type line: str
    :yield: key, match for key in :py:class:`ozi_spec.CommentPatterns` excluding ``help``
    :rtype: Generator[tuple[str, str], None, None]
    """

def diagnose(
    line: str, rel_path: Path, line_no: int
) -> Generator[tuple[str, str], None, None]:
    """Diagnose OZI comment pattern for a single line.

    :param line: line text verbatim
    :type line: str
    :param rel_path: file relative to OZI project root
    :type rel_path: Path
    :param line_no: current line number
    :type line_no: int
    """

def diagnostic(
    lines: Sequence[str], rel_path: Path, start: int = ...
) -> tuple[Counter[str], dict[str, str]]:
    """Diagnose OZI comment pattern for a sequence of lines (usually a single file).

    :param lines: lines to check
    :type lines: list[str]
    :param rel_path: file relative to OZI project root
    :type rel_path: Path
    :param start: starting line number for asynchronous chunking, defaults to 1
    :type start: int, optional
    :rtype: tuple[Counter[str], dict[str, str]]
    :return: count of lines and comment pattern matches as a dict
    """

def score_file(count: Counter[str]) -> float:
    """Score a single file comment diagnostic.

    .. deprecated:: 1.11.2

        Use :py:func:`ozi.comment.score` instead.
    """

def score(count: Counter[str]) -> float:
    """Score a single comment diagnostic.

    :param count: count of lines and comments
    :type count: Counter[str]
    :return: span comment score out of 5.0
    :rtype: float
    """

def comment_diagnostic(target: Path, rel_path: Path, file: str) -> None:
    """Run a scored comment diagnostic on a python file."""



# ozi/meson.py
# Part of the OZI Project, under the Apache License v2.0 with LLVM Exceptions.
# See LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
"""A query tool for meson.build files."""
from __future__ import annotations

from functools import lru_cache
from typing import TypeAlias

from mesonbuild.ast.interpreter import AstInterpreter
from mesonbuild.interpreterbase.exceptions import InvalidArguments
from mesonbuild.mparser import ArrayNode
from mesonbuild.mparser import AssignmentNode
from mesonbuild.mparser import BaseNode
from mesonbuild.mparser import CodeBlockNode
from mesonbuild.mparser import DictNode
from mesonbuild.mparser import ElementaryNode
from mesonbuild.mparser import ForeachClauseNode
from mesonbuild.mparser import FunctionNode
from mesonbuild.mparser import IdNode
from mesonbuild.mparser import MethodNode
from mesonbuild.mparser import NotNode
from mesonbuild.mparser import PlusAssignmentNode
from mesonbuild.mparser import StringNode
from mesonbuild.mparser import UMinusNode
from tap_producer import TAP

from ozi_core._i18n import TRANSLATION

SelectValue: TypeAlias = type[AssignmentNode | PlusAssignmentNode | NotNode | UMinusNode]
SelectItems: TypeAlias = type[ForeachClauseNode]
WhereValue: TypeAlias = type[ArrayNode | DictNode | MethodNode | FunctionNode]
WhereItems: TypeAlias = type[IdNode]


@lru_cache
def load_ast(source_root: str) -> CodeBlockNode | None:
    """Load the :abbr:`AST (Abstract Syntax Tree)` from the root :file:`meson.build`.

    :param source_root: Directory containing a top-level :file:`meson.build`.
    :type source_root: str
    :return: The AST for a meson build definition if one is available OR None.
    :rtype: CodeBlockNode | None
    """
    ast = AstInterpreter(source_root, '', '')  # pyright: ignore
    try:
        ast.load_root_meson_file()
    except InvalidArguments:  # pragma: no cover
        return None
    return ast.ast


def project_metadata(ast: CodeBlockNode) -> tuple[str, str]:
    """Extract project metadata from :file:`meson.build`

    :param ast: The AST for a :file:`meson.build`.
    :type ast: CodeBlockNode
    :return: The project name and license identifier.
    :rtype: tuple[str, str]
    """
    project_args = ast.lines[0].args.arguments  # pyright: ignore
    default_options = ast.lines[0].args.kwargs  # pyright: ignore
    license_ = [
        v
        for k, v in default_options.items()
        if isinstance(k, IdNode) and k.value == 'license'
    ][0]

    if isinstance(license_, ArrayNode):  # pragma: no cover
        license_ = license_.args.arguments[0]
        TAP.comment(
            TRANSLATION('term-tap-meson-build-license-array'),
            license_.value,  # pyright: ignore
        )

    license_ = license_.value  # pyright: ignore

    project_name, *_ = (i.value for i in project_args if isinstance(i, StringNode))

    return project_name, license_


@lru_cache(typed=True)
def query_simple(
    ast: CodeBlockNode,
    select: SelectValue = AssignmentNode,
    where: WhereValue = ArrayNode,
) -> set[BaseNode]:  # pragma: no cover
    """Run a simplistic query with no node filtering.

    :param ast: The AST for a :file:`meson.build`.
    :type ast: CodeBlockNode
    :param select: Select node type, defaults to AssignmentNode
    :type select: SelectValue, optional
    :param where: Where node type, defaults to ArrayNode
    :type where: WhereValue, optional
    :return: Set of all nodes matching the selector.
    :rtype: set[BaseNode]
    """
    matches, *_ = (
        (
            i.value.args.arguments
            for i in ast.lines
            if isinstance(i, select) and isinstance(i.value, where)
        )
        if ast
        else (_ for _ in ())
    )
    return set(matches)


@lru_cache(typed=True)
def query_complex(
    ast: CodeBlockNode,
    select: SelectValue = AssignmentNode,
    where: WhereValue = ArrayNode,
    get: type[ElementaryNode] = StringNode,
) -> set[str]:  # pragma: no cover
    """Run a complex (filtered) query of build item values.

    :param ast: The AST for a :file:`meson.build`.
    :type ast: CodeBlockNode
    :param select: select type, defaults to AssignmentNode
    :type select: SelectValue, optional
    :param where: where type, defaults to ArrayNode
    :type where: WhereValue, optional
    :param get: get type, defaults to StringNode
    :type get: type[ElementaryNode], optional
    :return: A set of node values matching the selector
    :rtype: set[str]
    """
    matches, *_ = (
        (
            i.value.args.arguments
            for i in ast.lines
            if isinstance(i, select) and isinstance(i.value, where)
        )
        if ast
        else (_ for _ in ())
    )
    return {i.value for i in matches if isinstance(i, get)}


@lru_cache(typed=True)
def query_loop_assignments(
    ast: CodeBlockNode,
    select: SelectItems = ForeachClauseNode,
    where: type[ElementaryNode] = IdNode,
    get: type[ElementaryNode] = StringNode,
) -> set[str]:  # pragma: no cover
    """Get a set of selected array items (must be assigned to a variable).

    :param ast: The AST for a :file:`meson.build`.
    :type ast: CodeBlockNode
    :param select: Select node type, defaults to ForeachClauseNode
    :type select: SelectItems, optional
    :param where: Where node type, defaults to IdNode
    :type where: WhereItems, optional
    :param get: Query node type, defaults to StringNode
    :type get: type[ElementaryNode], optional
    :return: A set of query matches
    :rtype: set[str]
    """
    loop_vars = (
        (
            i.items.value
            for i in ast.lines
            if isinstance(i, select) and isinstance(i.items, where)
        )
        if ast
        else (_ for _ in ())
    )
    assigned = (
        i.value.args.arguments
        for i in ast.lines
        if isinstance(i, AssignmentNode)
        and isinstance(i.value, ArrayNode)
        and i.var_name in set(loop_vars)
    )
    return {i.value for i in assigned if isinstance(i, get)}


@lru_cache(typed=True)
def find_var_suffix(
    ast: CodeBlockNode,
    var_suffix: str,
) -> set[str]:  # pragma: no cover
    """Get a set of build items based on a variable name suffix.

    :param ast: The AST for a :file:`meson.build`.
    :type ast: CodeBlockNode
    :param var_suffix: The text to look for.
    :type var_suffix: str
    :return: A set of query matches
    :rtype: set[str] | None
    """
    loop_vars = {
        i.items.value
        for i in ast.lines
        if isinstance(i, ForeachClauseNode)
        and isinstance(i.items, IdNode)
        and i.items.value.endswith(var_suffix)
    }
    assigned = [
        i.value.args.arguments
        for i in ast.lines
        if isinstance(i, AssignmentNode)
        and i.var_name in loop_vars
        and isinstance(i.value, ArrayNode)
    ]
    if len(assigned) > 0:  # pragma: defer to good-issue
        return {i.value for i in assigned[0] if isinstance(i, StringNode)}
    else:
        return set()


def query_build_value(
    source_root: str,
    query: str,
) -> bool | None:  # pragma: no cover
    """Load a :file:`meson.build` project and check if a query exists in array assignments.

    :param source_root: The path to directory containing a :file:`meson.build`
    :type source_root: str
    :param query: The text to look for.
    :type query: str
    :return: True if a query match is found, False if not, and None if the
             :file:`meson.build` could not be loaded.
    :rtype: set[str] | None
    """
    ast = load_ast(source_root)
    if ast:
        build_data = query_complex(ast)
        return query in build_data
    return ast


def get_items_by_suffix(
    source_root: str,
    query: str,
) -> set[str] | None:  # pragma: no cover
    """Load a :file:`meson.build` project and return build items.

    :param source_root: The path to directory containing a :file:`meson.build`
    :type source_root: str
    :param query: The text to look for.
    :type query: str
    """
    ast = load_ast(source_root)
    if ast:
        return find_var_suffix(ast, query)
    return ast



from __future__ import annotations

from enum import Enum
from enum import auto
from enum import unique

from ozi_core._i18n import TRANSLATION


def checkbox(checked: bool) -> str:  # pragma: no cover
    if checked:
        return '☑'
    else:
        return '☐'


@unique
class MenuButton(Enum):
    """Non-composable menu action enum."""

    ADD = auto()
    BACK = auto()
    EDIT = auto()
    EXIT = auto()
    MENU = auto()
    METADATA = auto()
    OK = auto()
    OPTIONS = auto()
    PROMPT = auto()
    REMOVE = auto()
    RESET = auto()
    YES = auto()
    NO = auto()
    SKIP = auto()

    @property
    def _tuple(self: MenuButton) -> tuple[str, int]:  # pragma: no cover
        """Return a tuple of text, value for prompt-toolkit buttons."""
        return TRANSLATION(f'btn-{self.name.lower()}'), self.value

    @property
    def _str(self: MenuButton) -> str:  # pragma: no cover
        """Return prompt-toolkit button text."""
        return TRANSLATION(f'btn-{self.name.lower()}')
